/*============================================================================
	File:		0020 - PostgreSQL - Concurrency.sql

	Summary:	This script demonstrates how PostgreSQL
				is handling concurrency.

	Date:		September 2023

	PostgreSQL Version: 14 / 15
------------------------------------------------------------------------------
	Written by Uwe Ricken, db Berater GmbH

	This script is intended only as a supplement to demos and lectures
	given by Uwe Ricken.  
  
	THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
	ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
	TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
	PARTICULAR PURPOSE.
============================================================================*/
DROP TABLE IF EXISTS "dbo"."Customers";
CREATE TABLE "dbo"."Customers"
(
	Id			INT				NOT NULL	PRIMARY KEY,
	name		VARCHAR(256)	NOT NULL,
	InsertUser	VARCHAR(128)	NOT NULL	DEFAULT (user),
	InsertDate	TIMESTAMP		NOT NULL	DEFAULT (NOW())
);



/* A typical workload when it comes to changes in records */
BEGIN TRANSACTION UpdateCustomer;

	UPDATE	"dbo"."Customers"
	SET		name = 'SQL Konferenz 2023'
	WHERE	Id = 10;

	/*
		Run the following command in a new session!

		SELECT	Id,
				Name,
				InsertUser,
				InsertDate
		FROM	dbo.Customers
		WHERE	Id = 10;
	*/

	/* When the second session is running we check the conflicts */
	;WITH SortOrder
	AS
	(
		SELECT	SortOrderKey,
				resource_type
		FROM	(
				VALUES (1, N'OBJECT'),
						(2, N'PAGE'),
						(3, N'KEY'),
						(3, N'RID')
				) AS R (SortOrderKey, resource_type)
	)
	SELECT	dtl.resource_type,
            dtl.resource_description,
            dtl.request_mode,
            dtl.request_type,
            dtl.request_status,
            dtl.request_session_id
	FROM	sys.dm_tran_locks AS dtl
			INNER JOIN SortOrder AS so
			ON (dtl.resource_type = so.resource_type)
	WHERE	dtl.resource_type <> N'DATABASE'
			AND resource_database_id = DB_ID()
	ORDER BY
			dtl.request_session_id,
			so.SortOrderKey;
	GO

ROLLBACK TRANSACTION;
GO

/*
	What possibilities do we have to avoid locking?

	NOLOCK:				DON'T USE THIS SHIT!
	READUNCOMMITTED:	is exactly the same as NOLOCK!

	RCSI:				optimistic locking
*/
/* NOLOCK */
SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (NOLOCK)
WHERE	Id = 10;
GO

/* READUNCOMMITTED */
SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (READUNCOMMITTED)
WHERE	Id = 10;
GO

/* READ COMMITTED SNAPSHOT ISOLATION */
USE master;
GO

ALTER DATABASE demo_db SET READ_COMMITTED_SNAPSHOT ON WITH ROLLBACK IMMEDIATE;
GO

USE demo_db;
GO

SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (READUNCOMMITTED)
WHERE	Id = 10;
GO
