/*============================================================================
	File:		0020 - SQL Server - Concurrency.sql

	Summary:	This script demonstrates how Microsoft SQL Server
				is handling concurrency.

	Date:		September 2023

	SQL Server Version: >= 2016
------------------------------------------------------------------------------
	Written by Uwe Ricken, db Berater GmbH

	This script is intended only as a supplement to demos and lectures
	given by Uwe Ricken.  
  
	THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
	ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
	TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
	PARTICULAR PURPOSE.
============================================================================*/
USE master;
GO

/* Create a demo database first */
EXEC dbo.sp_create_demo_db;
GO

USE demo_db;
GO

SELECT * INTO dbo.Customers FROM CustomerOrders.dbo.Customers;
GO

/*
	Keep in mind that only Microsoft SQL Server is using
	Clustered Index Structures!	
*/
ALTER TABLE dbo.Customers
ADD CONSTRAINT pk_Customers PRIMARY KEY CLUSTERED (Id);
GO

/* A typical workload when it comes to changes in records */
BEGIN TRANSACTION UpdateCustomer;
GO
	UPDATE	dbo.Customers
	SET		name = 'SQL Konferenz 2023'
	WHERE	Id = 10;
	GO

	/*
		Run the following command in a new session!

		SELECT	Id,
				Name,
				InsertUser,
				InsertDate
		FROM	dbo.Customers
		WHERE	Id = 10;
	*/

	/* When the second session is running we check the conflicts */
	;WITH SortOrder
	AS
	(
		SELECT	SortOrderKey,
				resource_type
		FROM	(
				VALUES (1, N'OBJECT'),
						(2, N'PAGE'),
						(3, N'KEY'),
						(3, N'RID')
				) AS R (SortOrderKey, resource_type)
	)
	SELECT	dtl.resource_type,
            dtl.resource_description,
            dtl.request_mode,
            dtl.request_type,
            dtl.request_status,
            dtl.request_session_id
	FROM	sys.dm_tran_locks AS dtl
			INNER JOIN SortOrder AS so
			ON (dtl.resource_type = so.resource_type)
	WHERE	dtl.resource_type <> N'DATABASE'
			AND resource_database_id = DB_ID()
	ORDER BY
			dtl.request_session_id,
			so.SortOrderKey;
	GO

ROLLBACK TRANSACTION;
GO

/*
	What possibilities do we have to avoid locking?

	NOLOCK:				DON'T USE THIS SHIT!
	READUNCOMMITTED:	is exactly the same as NOLOCK!

	RCSI:				optimistic locking
*/
/* NOLOCK */
SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (NOLOCK)
WHERE	Id = 10;
GO

/* READUNCOMMITTED */
SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (READUNCOMMITTED)
WHERE	Id = 10;
GO

/* READ COMMITTED SNAPSHOT ISOLATION */
USE master;
GO

ALTER DATABASE demo_db SET READ_COMMITTED_SNAPSHOT ON WITH ROLLBACK IMMEDIATE;
GO

USE demo_db;
GO

SELECT	Id,
		Name,
		InsertUser,
		InsertDate
FROM	dbo.Customers WITH (READUNCOMMITTED)
WHERE	Id = 10;
GO
