/*============================================================================
	File:		0040 - SQL Server - Statistics.sql

	Summary:	This script demonstrates the behavior of a query when it detects
				outdated statistics.

	Date:		September 2023

	SQL Server Version: 2016 / 2017 / 2019 / 2022
------------------------------------------------------------------------------
	Written by Uwe Ricken, db Berater GmbH

	This script is intended only as a supplement to demos and lectures
	given by Uwe Ricken.  
  
	THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
	ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
	TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
	PARTICULAR PURPOSE.
============================================================================*/
USE master;
GO

EXEC sp_create_demo_db;
GO

USE demo_db;
GO


/* Let's create a demo table with a few records */
SELECT	message_id,
        language_id,
        severity,
        is_event_logged,
        text
INTO	dbo.messages
FROM	sys.messages
WHERE	language_id = 1031;
GO

/* Now we create an index on the severity column and check the statistics */
CREATE NONCLUSTERED INDEX nix_messages_severity ON dbo.messages (severity);
GO

/* what statistics do we have and what is the modification counter? */
SELECT	S.name,
		SP.object_id,
        SP.stats_id,
        SP.last_updated,
        SP.rows,
        SP.rows_sampled,
        SP.steps,
        SP.unfiltered_rows,
        SP.modification_counter,
        SP.persisted_sample_percent
FROM	sys.stats AS S
		CROSS APPLY sys.dm_db_stats_properties
		(
			S.object_id,
			S.stats_id
		) AS SP
WHERE	s.object_id = OBJECT_ID(N'dbo.messages', N'U');
GO

SET STATISTICS IO ON;
GO

/* let's run a query against the table to check the statistics */
SELECT * FROM dbo.messages WHERE severity = 13;
GO

SELECT * FROM dbo.messages WHERE severity = 16
GO

/*
	we have ~15,000 records in the table (depends on the version of SQL Server!)
	20% + 500 changes = 3,000 + 500 =>	3,500 changes!
*/

/* Let's enter 2.000 more records into dbo.messages */
;WITH m
AS
(
	SELECT	message_id,
            language_id,
            severity,
            is_event_logged,
            text
	FROM	sys.messages
	WHERE	language_id = 1033
)
INSERT INTO dbo.messages WITH (TABLOCK)
(
	message_id,
	language_id,
	severity,
	is_event_logged,
	[text]
)
SELECT	TOP (2000)
		message_id, language_id, severity, is_event_logged, [text]
FROM	m

EXCEPT

SELECT	message_id,
        language_id,
        severity,
        is_event_logged,
        text
FROM	dbo.messages;
GO

/* what statistics do we have and what is the modification counter? */
SELECT	S.name,
		SP.object_id,
        SP.stats_id,
        SP.last_updated,
        SP.rows,
        SP.rows_sampled,
        SP.steps,
        SP.unfiltered_rows,
        SP.modification_counter,
        SP.persisted_sample_percent
FROM	sys.stats AS S
		CROSS APPLY sys.dm_db_stats_properties
		(
			S.object_id,
			S.stats_id
		) AS SP
WHERE	s.object_id = OBJECT_ID(N'dbo.messages', N'U');
GO

/* let's run a query against the table to check the statistics */
SELECT	message_id,
		language_id,
		severity,
		is_event_logged,
		text
FROM	dbo.messages
WHERE	severity = 13
OPTION (USE HINT ('FORCE_LEGACY_CARDINALITY_ESTIMATION'));
GO

SELECT	message_id,
		language_id,
		severity,
		is_event_logged,
		text
FROM	dbo.messages
WHERE	severity = 13
OPTION (USE HINT ('FORCE_DEFAULT_CARDINALITY_ESTIMATION'));
GO

SELECT	message_id,
		language_id,
		severity,
		is_event_logged,
		text
FROM	dbo.messages
WHERE	severity = 16
OPTION (USE HINT ('FORCE_LEGACY_CARDINALITY_ESTIMATION'));
GO

SELECT	message_id,
		language_id,
		severity,
		is_event_logged,
		text
FROM	dbo.messages
WHERE	severity = 16
OPTION (USE HINT ('FORCE_DEFAULT_CARDINALITY_ESTIMATION'));
GO

/* Let's enter another 2.000 more records into dbo.messages */
;WITH m
AS
(
	SELECT	message_id,
            language_id,
            severity,
            is_event_logged,
            text
	FROM	sys.messages
	WHERE	language_id = 1028
)
INSERT INTO dbo.messages WITH (TABLOCK)
(
	message_id,
	language_id,
	severity,
	is_event_logged,
	[text]
)
SELECT	TOP (2000)
		message_id, language_id, severity, is_event_logged, [text]
FROM	m

EXCEPT

SELECT	message_id,
        language_id,
        severity,
        is_event_logged,
        text
FROM	dbo.messages;
GO

-- what statistics do we have and what is the modification counter?
SELECT	S.name,
		SP.object_id,
        SP.stats_id,
        SP.last_updated,
        SP.rows,
        SP.rows_sampled,
        SP.steps,
        SP.unfiltered_rows,
        SP.modification_counter,
        SP.persisted_sample_percent
FROM	sys.stats AS S
		CROSS APPLY sys.dm_db_stats_properties
		(
			S.object_id,
			S.stats_id
		) AS SP
WHERE	s.object_id = OBJECT_ID(N'dbo.messages', N'U');
GO

DBCC SHOW_STATISTICS(N'dbo.messages', N'nix_messages_severity')
WITH STAT_HEADER;
DBCC SHOW_STATISTICS(N'dbo.messages', N'nix_messages_severity')
WITH HISTOGRAM
GO

-- let's run a query against the table to check the statistics
EXEC sp_executesql N'SELECT * FROM dbo.messages WHERE severity = @Id;',
N'@Id TINYINT', 13;
GO

ALTER DATABASE SCOPED CONFIGURATION CLEAR PROCEDURE_CACHE;
GO

EXEC sp_executesql N'SELECT * FROM dbo.messages WHERE severity = @Id;',
N'@Id TINYINT', 16;
GO

UPDATE STATISTICS dbo.messages WITH FULLSCAN;
GO

-- what statistics do we have and what is the modification counter?
SELECT	S.name,
		SP.object_id,
        SP.stats_id,
        SP.last_updated,
        SP.rows,
        SP.rows_sampled,
        SP.steps,
        SP.unfiltered_rows,
        SP.modification_counter,
        SP.persisted_sample_percent
FROM	sys.stats AS S
		CROSS APPLY sys.dm_db_stats_properties
		(
			S.object_id,
			S.stats_id
		) AS SP
WHERE	s.object_id = OBJECT_ID(N'dbo.messages', N'U');
GO